
#include <stdio.h>
#include <stdlib.h>

#include "common.h"
#include "sdb.h"

#define WL "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz,-/"

static int db_callback(int n, char **p, void *closure)
{
	int i;
	for (i = 0; i < n; i++) {
		printf("%s ", p[i]);
	}
	printf("\n");
	return 0;
}

static void usage(void)
{
	fprintf(stderr, "usage: ucronadm args...\n");
	exit(EXIT_FAILURE);
}

static int add(char **argv)
{
	char *min = NULL, *hour = NULL, *day = NULL;
	char *month = NULL, *dayofweek = NULL;
	char *command = NULL, *hostname = NULL, *username = NULL;
	char b[4096];
	int n;
	int serial;

	if (!(min = argv[2]) || !(hour = argv[3]) || !(day = argv[4]) ||
	    !(month = argv[5]) || !(dayofweek = argv[6]) ||
	    !(command = argv[7]) || !(hostname = argv[8])) {
		usage();
	}
	whitelist(min, WL, 0);
	whitelist(hour, WL, 0);
	whitelist(day, WL, 0);
	whitelist(month, WL, 0);
	whitelist(dayofweek, WL, 0);

	if (!(username = argv[9])) username = "root";
	serial = counter("serial");
	snprintf(b, sizeof b,
		"insert into crontab ("
			"min, hour, day,"
			"month, dayofweek, command,"
			"hostname, username, serial) "
		"values ("
			"'%s', '%s', '%s',"
			"'%s', '%s', '%s',"
			"'%s', '%s', %d)",
		min, hour, day,
		month, dayofweek, command,
		hostname, username, serial);
	n = sdb_query(url, b, db_callback, NULL);
	return n;
}

static int list(char **argv)
{
	char *hostname = NULL, *username = NULL;
	char b[4096];
	int n;

	if (!(hostname = argv[2])) {
		usage();
	}
	if (!(username = argv[3])) username = "root";
	snprintf(b, sizeof b,
		"select min, hour, day, month, dayofweek, command "
		"from crontab "
		"where hostname='%s' "
		"and username='%s'",
		hostname, username);
	n = sdb_query(url, b, db_callback, NULL);
	return n;
}

static int Delete(char **argv)
{
	char *hostname = NULL, *username = NULL;
	char b[4096];
	int n;

	if (!(hostname = argv[2])) {
		fprintf(stderr, "not implemented\n");
		exit(EXIT_FAILURE);
	}
	if (!(username = argv[3])) username = "root";
	snprintf(b, sizeof b,
		"delete from crontab "
		"where hostname='%s' "
		"and username='%s'",
		hostname, username);
	n = sdb_query(url, b, db_callback, NULL);
	return n;
}

static int db_upload(int n, char **p, void *closure)
{
	int i;
	FILE *fp_up = (FILE *)closure;

	for (i = 0; i < n; i++) {
		fprintf(fp_up, "%s ", p[i]);
	}
	fprintf(fp_up, "\n");
	return 0;
}

static int upload(char **argv)
{
	char *hostname = NULL, *username = NULL;
	char b[4096];
	int n;
	FILE *fp_up;

	if (!(hostname = argv[2])) {
		usage();
	}
	if (!(username = argv[3])) username = "root";
	snprintf(b, sizeof b, "%s upload %s %s", USH, hostname, username);
	fp_up = popen(b, "w");
	if (!fp_up) return -1;
	snprintf(b, sizeof b,
		"select min, hour, day, month, dayofweek, command "
		"from crontab "
		"where hostname='%s' "
		"and username='%s'",
		hostname, username);
	n = sdb_query(url, b, db_upload, (void *)fp_up);
	pclose(fp_up);
	return n;
}

static int download(char **argv)
{
	char *hostname = NULL, *username = NULL;
	char b[4096];
	char min[100], hour[100], day[100], month[100], dayofweek[100];
	char command[1000];
	FILE *fp;
	int n;
	int serial;

	if (!(hostname = argv[2])) {
		usage();
	}
	if (!(username = argv[3])) username = "root";

	serial = get_counter("serial")+1;
	snprintf(b, sizeof b, "%s download %s %s", USH, hostname, username);
	fp = popen(b, "r");
	if (!fp) return -1;
	while (fgets(b, sizeof b, fp)) {
		chomp(b);
		n = sscanf(b, "%s %s %s %s %s %[^\n]",
			   min, hour, day, month, dayofweek, command);
		if (n == 6) {
			whitelist(min, WL, 0);
			whitelist(hour, WL, 0);
			whitelist(day, WL, 0);
			whitelist(month, WL, 0);
			whitelist(dayofweek, WL, 0);
			snprintf(b, sizeof b,
				"insert into crontab ("
					"min, hour, day,"
					"month, dayofweek, command,"
					"hostname, username, serial) "
				"values ("
					"'%s', '%s', '%s',"
					"'%s', '%s', '%s',"
					"'%s', '%s', %d)",
				min, hour, day, month, dayofweek,
				command, hostname, username, serial++);
			n = sdb_query(url, b, db_callback, NULL);
		}
	}
	put_counter("serial", serial);
	pclose(fp);
	return 0;
}

static int script(char **argv)
{
	char *hostname = NULL, *local = NULL, *remote = NULL;
	char b[4096];

	if (!(hostname = argv[2]) || \
	    !(local = argv[3]) || \
	    !(remote = argv[4])) {
printf("hostname=%s, local=%s, remote=%s\n", hostname, local, remote);
		usage();
	}
	snprintf(b, sizeof b, "%s script %s %s %s",
		USH, hostname, local, remote);
	system(b);
	return 0;
}

static int delscript(char **argv)
{
	char *hostname = NULL, *filename = NULL;
	char b[4096];

	if (!(hostname = argv[2]) || !(filename = argv[3])) {
		usage();
	}
	snprintf(b, sizeof b, "%s delscript %s %s", USH, hostname, filename);
	system(b);
	return 0;
}

static int ls(char **argv)
{
	char *hostname = NULL;
	FILE *fp;
	char b[4096];

	if (!(hostname = argv[2])) {
		usage();
	}
	snprintf(b, sizeof b, "%s ls %s", USH, hostname);
	fp = popen(b, "r");
	if (!fp) return -1;
	while (fgets(b, sizeof b, fp)) {
		chomp(b);
		printf("%s\n", b);
	}
	pclose(fp);
	return 0;
}

static struct {
	char *name;
	int (*fn)(char **);
} fns[] = {
	{"add", add},
	{"list", list},
	{"delete", Delete},
	{"upload", upload},
	{"download", download},
	{"script", script},
	{"delscript", delscript},
	{"ls", ls},
	{NULL, NULL}};

int main(int argc, char **argv)
{
	char *p = argv[1];
	int i, n;

	sdb_init();
	url = getenv("UCRON_URL");
	if (!url) url = UCRON_URL;

	if (p == NULL) {
		usage();
	}
	if (p[0] == '!') {
		n = sdb_query(url, p+1, db_callback, NULL);
		exit(0);
	}
	for (i = 0; fns[i].name; i++) {
		if (!strcmp(p, fns[i].name)) break;
	}
	if (fns[i].name) {
		n = (*fns[i].fn)(argv);
	} else {
		usage();
	}

	return 0;
}

