
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "common.h"
#include "sdb.h"

#define TMPFILE "/tmp/ucroncgi.tmp"
#define WL "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz,-/*"

struct {
	char *attr;
	char *value;
} query[1000];
int nquery;

char *query_string;

/* old-fashioned html generating code */

/* translate the %NN stuff used in URLs into earthling characters */
static void decode_url(char *p)
{
        char *q = p;
        int c;

        while ((c = *p++)) {
                if (c == '+') {
                        *q++ = ' ';
                } else if (c == '%') {
                        sscanf(p, "%2x", &c);
                        *q++ = c;
                        if (*p) p++;
                        if (*p) p++;
                } else {
                        *q++ = c;
                }
        }
        *q = '\0';
}

static char *cgi_err = "none";

/* returns -1 for failure, 0 for success */
static int parsecgi(void)
{
	char *method = getenv("REQUEST_METHOD");
	char *type = getenv("CONTENT_TYPE");
	char *p;
	int i;

	nquery = 0;
	if (method == NULL) {
		cgi_err = "parsecgi: no method";
		return -1;
	}
	upcase(method);
	if (!strcmp(method, "GET")) {
		char *qs = getenv("QUERY_STRING");
		if (qs == NULL) {
			cgi_err = "parsecgi: no QUERY_STRING in GET";
			return -1;
		}
		query_string = cstrdup(qs);
	} else if (!strcmp(method, "POST")) {
		int n;
		char *ql = getenv("CONTENT_LENGTH");
		if (type && !strncmp(type, "multipart/", 10)) {
			/* mimecgi() */
			cgi_err = "parsecgi: multipart";
			return -1;	/* we don't handle this yet */
		}
		if (ql == NULL) {
			cgi_err = "parsecgi: no CONTENT_LENGTH";
			return -1;
		}
		n = atoi(ql);
		query_string = cmalloc(n+1);
		fread(query_string, 1, n, stdin);
		query_string[n] = '\0';
	} else {
		cgi_err = "parsecgi: bogus method";
		return -1;
	}
	for (p = strtok(query_string, "&"); p; p = strtok(NULL, "&")) {
		query[nquery++].attr = p;
	}
	for (i = 0; i < nquery; i++) {
		p = strchr(query[i].attr, '=');
		if (p) {
			*p = '\0';
			query[i].value = p+1;
			decode_url(query[i].value);
		} else {
			query[i].value = NULL;
		}
	}
	return 0;
}

/* returns value for success, NULL for failure */
static char *get_query(char *attr)
{
        int i;
        for (i = 0; i < nquery; i++) {
                if (!strcmp(attr, query[i].attr)) {
			return query[i].value;
		}
	}
        return NULL;
}

static void html_header(char *meta, char *title)
{
	char *hostname = get_query("hostname");
	char *username = get_query("username");
	printf("Content-type: text/html\n\n"
		"<html>\n"
		"<head>\n");
	if (meta) printf("%s\n", meta);
	printf("<title>%s</title>\n", title);
	printf("</head>\n\n"
		"<body link=\"#ffffff\" vlink=\"#ffffff\" alink=\"#ffffff\" "
		"text=\"#ffffff\" bgcolor=\"#001C43\">\n"
		"<!-- Outer table -->\n"
		"<table width=\"100%%\" border=0 cellpadding=0 cellspacing=5>\n"
		"<tr>\n"
		"<td width=16>\n"
		"</td>\n"
		"<td>\n"
		"<!-- Logo here -->\n"
		"<img src=\"/logo.gif\" alt=\"\">\n"
		"<hr size=1 noshade>\n\n"
		"<!-- Inner table -->\n"
		"<table width=\"100%%\">\n"
		"<tr>\n"
		"<td valign=top width=130>\n"
		"<!-- Left pane -->\n");
	printf("<a href=\"%s\"><b>Home</b></a><p>", SCRIPT);
	if (hostname && hostname[0]) {
		if (!username || !username[0]) username = "root";
		printf("<a href=\"%s?mode=Crontab&hostname=%s&username=%s\">"
			"<b>Crontab</b></a><p>",
			SCRIPT, hostname, username);
		printf("<a href=\"%s?mode=Scripts&hostname=%s\">"
			"<b>Scripts</b></a><p>",
			SCRIPT, hostname);
	}
	printf("</td>\n"
		"<td valign=top width=8 bgcolor=\"#cccccc\">\n"
		"&nbsp;\n"
		"</td>\n"
		"<td valign=top bgcolor=\"#001C43\">\n");
}

static void html_footer(void)
{
	printf("<!-- Footer -->\n"
		"&nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;\n"
		"&nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;\n"
		"&nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;\n"
		"&nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;\n"
		"&nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;\n"
		"&nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;\n"
		"&nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;\n"
		"&nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;\n"
		"&nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;\n"
		"&nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;\n"
		"&nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;\n"
		"&nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;\n"
		"&nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;\n"
		"&nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;\n"
		"&nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;\n"
		"&nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;\n"
		"&nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;\n"
		"</td>\n"
		"</tr>\n"
		"</table>\n\n"
		"<hr size=1 noshade>\n"
		"<p>\n\n"
		"<!-- Insert contact info here -->\n"
		"</td>\n"
		"</tr>\n"
		"</table>\n\n"
		"</body>\n"
		"</html>\n");
}

static void html_error(char *p)
{
	html_header(NULL, "Error");
	printf("<h1>Error</h1>\n");
	printf("Reason: %s<p>\n", p);
	printf("cgi_err: %s\n", cgi_err);
	html_footer();
}

static char *require_query(char *attr)
{
	char *p = get_query(attr);
	char b[4096];

	if (!p || !p[0]) {
		snprintf(b, sizeof b, "Missing required: %s", attr);
		html_error(b);
		exit(0);
	}
	return p;
}

static void html_default(void)
{
	html_header(NULL, "Main menu");
	printf("<h1>Main menu</h1>\n");
	printf("<form method=get action=\"%s\">\n", SCRIPT);
	printf("<table>\n"
		"<tr><td><b>Hostname:</b></td>"
		"<td><input type=text size=65 name=hostname value=\"\"></td>"
		"</tr>\n"
		"<tr><td><b>Username:</b></td>"
		"<td><input type=text size=65 name=username value=\"\"></td>"
		"</tr>\n"
		"</table>\n");
	printf("<input type=submit name=mode value=\"Crontab\">\n");
	printf("<input type=submit name=mode value=\"Scripts\">\n");
	printf("</form>\n");
	html_footer();
}

struct listinfo {
	char *hostname;
	char *username;
};

static void jobrow(char *min, char *hour, char *day,
		char *month, char *dayofweek, char *command,
		char *hostname, char *username, char *serial)
{
	printf("<form>\n");
	printf("<tr>\n");
	printf("<td><input type=text size=5 name=min value=\"%s\"></td>\n"
		"<td><input type=text size=5 name=hour value=\"%s\"></td>\n"
		"<td><input type=text size=5 name=day value=\"%s\"></td>\n"
		"<td><input type=text size=5 name=month value=\"%s\"></td>\n"
		"<td><input type=text size=5 name=dayofweek value=\"%s\"></td>\n"
		"<td><input type=text name=command value=\"%s\"></td>\n",
		min, hour, day, month, dayofweek, command);
	printf("<input type=hidden name=hostname value=\"%s\">\n"
		"<input type=hidden name=username value=\"%s\">\n",
		hostname, username);
	printf("<td><input type=submit name=mode value=\"Add\">\n");
	if (serial) {
		printf("<input type=hidden name=serial value=\"%s\">\n"
			"<input type=submit name=mode value=\"Update\">\n"
			"<input type=submit name=mode value=\"Delete\">"
			"</td>\n",
			serial);
	}
	printf("</tr>\n");
	printf("</form>\n");
}

static int db_list(int n, char **p, void *closure)
{
	struct listinfo *li = (struct listinfo *)closure;
	if (n < 7) return -1;
	jobrow(p[0], p[1], p[2], p[3], p[4], p[5],
		li->hostname, li->username, p[6]);
	return 0;
}

static void html_list(void)
{
	struct listinfo li;
	char *hostname = require_query("hostname");
	char *username = get_query("username");
	char b[4096];
	int n;

	if (!username || !username[0]) username = "root";
	li.hostname = hostname;
	li.username = username;
	snprintf(b, sizeof b,
		"select min, hour, day, month, dayofweek, command, serial "
		"from crontab "
		"where hostname='%s' "
		"and username='%s'",
		hostname, username);
	html_header(NULL, "List");
	printf("<h1>List for %s@%s</h1>\n", username, hostname);

#if 0	/* looks wrong */
	printf("<input type=hidden name=hostname value=\"%s\">\n"
		"<input type=hidden name=username value=\"%s\">\n",
		hostname, username);
#endif

	printf("<table>\n"
		"<tr>"
		"<td><b>Min</b></td>"
		"<td><b>Hour</b></td>"
		"<td><b>Day</b></td>"
		"<td><b>Month</b></td>"
		"<td><b>Weekday</b></td>"
		"<td><b>Command</b></td>"
		"<dt>&nbsp;</dt>"
		"</tr>\n");
	n = sdb_query(url, b, db_list, (void *)&li);

printf("<!-- Leaping Lizards -->\n");
	jobrow("", "", "", "", "", "", hostname, username, NULL);
printf("<!-- Holy Batman! -->\n");

	printf("</table>\n");
	printf("<form>\n");
	printf("<input type=hidden name=hostname value=\"%s\">\n", hostname);
	printf("<input type=hidden name=username value=\"%s\">\n", username);
	printf("<input type=submit name=mode value=\"Delete all\">\n");
	printf("<input type=submit name=mode value=\"Upload\">\n");
	printf("<input type=submit name=mode value=\"Download\">\n");
	printf("</form>\n");
	if (n == -1) {
		printf("DB error\n");
		printf("url='%s'\n", url);
	}
	html_footer();
}

static void html_addjob(void)
{
	char *min = require_query("min");
	char *hour = require_query("hour");
	char *day = require_query("day");
	char *month = require_query("month");
	char *dayofweek = require_query("dayofweek");
	char *command = require_query("command");
	char *hostname = require_query("hostname");
	char *username = require_query("username");
	char b[4096];
	int n;
	int serial;

	whitelist(min, WL, 0);
        whitelist(hour, WL, 0);
        whitelist(day, WL, 0);
        whitelist(month, WL, 0);
        whitelist(dayofweek, WL, 0);

	if (!username || !username[0]) username = "root";
	serial = counter("serial");
	snprintf(b, sizeof b,
		"insert into crontab ("
                        "min, hour, day,"
                        "month, dayofweek, command,"
                        "hostname, username, serial) "
                "values ("
                        "'%s', '%s', '%s',"
                        "'%s', '%s', '%s',"
                        "'%s', '%s', %d)",
                min, hour, day,
                month, dayofweek, command,
                hostname, username, serial);
        n = sdb_query(url, b, db_null, NULL);
	html_list();
}

static void html_update(void)
{
	char *min = require_query("min");
	char *hour = require_query("hour");
	char *day = require_query("day");
	char *month = require_query("month");
	char *dayofweek = require_query("dayofweek");
	char *command = require_query("command");
	char *username = require_query("username");
	char *serial = require_query("serial");
	char b[4096];
	int n;

	whitelist(min, WL, 0);
        whitelist(hour, WL, 0);
        whitelist(day, WL, 0);
        whitelist(month, WL, 0);
        whitelist(dayofweek, WL, 0);

	if (!username || !username[0]) username = "root";
	snprintf(b, sizeof b,
		"update crontab set "
			"min='%s',"
			"hour='%s',"
			"day='%s',"
			"month='%s',"
			"dayofweek='%s',"
			"command='%s' "
		"where serial=%s",
                min, hour, day, month, dayofweek, command, serial);
        n = sdb_query(url, b, db_null, NULL);
	html_list();
}

static void html_delall(void)
{
	char *hostname = require_query("hostname");
	char *username = get_query("username");
	char b[4096];
	int n;

	if (!username || !username[0]) username = "root";
	snprintf(b, sizeof b,
		"delete from crontab "
		"where hostname='%s' "
		"and username='%s'",
		hostname, username);
	n = sdb_query(url, b, db_null, NULL);
	html_list();
}

static void html_delone(void)
{
	char b[4096];
	char *serial = require_query("serial");
	snprintf(b, sizeof b, "delete from crontab where serial=%s", serial);
	sdb_query(url, b, db_null, NULL);
	html_list();
}

static int db_upload(int n, char **p, void *closure)
{
	int i;
	FILE *fp_up = (FILE *)closure;

	for (i = 0; i < n; i++) {
		fprintf(fp_up, "%s ", p[i]);
	}
	fprintf(fp_up, "\n");
	return 0;
}

static void html_upload(void)
{
	char *hostname = require_query("hostname");
	char *username = get_query("username");
	char b[4096];
	int n;
	FILE *fp_up;

	if (!username || !username[0]) username = "root";
	snprintf(b, sizeof b, "%s upload %s %s", USH, hostname, username);
	fp_up = popen(b, "w");
	if (!fp_up) {
		html_error("popen");
		return;
	}
	snprintf(b, sizeof b,
		"select min, hour, day, month, dayofweek, command "
		"from crontab "
		"where hostname='%s' "
		"and username='%s'",
		hostname, username);
	n = sdb_query(url, b, db_upload, (void *)fp_up);
	fflush(fp_up);
	pclose(fp_up);

/* This is a mystery. If we just do html_list now, which seems like the
   right thing to do, netscape gets a truncated file. It does not happen
   if we run this program from the command line.

   This only happens for html_upload. As a workaround, we create this
   simple page instead with a link back to the crontab listing.
*/
#if 0
	html_list();
#else
	snprintf(b, sizeof b,
		"<META HTTP-EQUIV=\"refresh\" "
		"content=\"1;URL=%s?mode=%s&hostname=%s&username=%s\">",
		SCRIPT, "Crontab", hostname, username);
	html_header(b, "Upload");
	printf("<h1>Upload complete</h1>\n");
	printf("<a href=\"%s?mode=Crontab&hostname=%s&username=%s\">Back</a>\n",
		SCRIPT, hostname, username);
	html_footer();
printf("<!-- %s -->\n", b);
#endif
}

static void html_download(void)
{
	char *hostname = require_query("hostname");
	char *username = get_query("username");
	char b[4096];
	char min[100], hour[100], day[100], month[100], dayofweek[100];
	char command[1000];
	FILE *fp;
	int n;
	int serial;

	if (!username || !username[0]) username = "root";

	serial = get_counter("serial")+1;
	snprintf(b, sizeof b, "%s download %s %s", USH, hostname, username);
	fp = popen(b, "r");
	if (!fp) {
		html_error("popen");
		return;
	}
	while (fgets(b, sizeof b, fp)) {
		chomp(b);
		n = sscanf(b, "%s %s %s %s %s %[^\n]",
			min, hour, day, month, dayofweek, command);
		if (n == 6) {
			whitelist(min, WL, 0);
        		whitelist(hour, WL, 0);
        		whitelist(day, WL, 0);
        		whitelist(month, WL, 0);
        		whitelist(dayofweek, WL, 0);
			snprintf(b, sizeof b,
				"insert into crontab ("
					"min, hour, day,"
					"month, dayofweek, command,"
					"hostname, username, serial) "
				"values ("
					"'%s', '%s', '%s',"
					"'%s', '%s', '%s',"
					"'%s', '%s', %d)",
				min, hour, day, month, dayofweek,
				command, hostname, username, serial++);
			n = sdb_query(url, b, db_null, NULL);
		}
	}
	put_counter("serial", serial);
	pclose(fp);
	html_list();
}

static void html_script(void)
{
	char *hostname = require_query("hostname");
	char *username = get_query("username");
	FILE *fp;
	char b[4096];

	if (!username || !username[0]) username = "root";
	snprintf(b, sizeof b, "%s ls %s", USH, hostname);
	fp = popen(b, "r");
	if (!fp) {
		html_error(USH);
		return;
	}
	html_header(NULL, "Scripts");
	printf("<h1>Scripts on %s</h1>\n", hostname);
	printf("<table>\n");
	while (fgets(b, sizeof b, fp)) {
		chomp(b);
		printf("<tr>\n"
			"<td>%s</td>\n", b);
		printf("<td>"
			"<a href=\"%s?mode=delscript&hostname=%s&"
			"username=%s&filename=%s\">"
			"Delete</a></td>\n"
			"</tr>\n",
			SCRIPT, hostname, username, b);
	}
	pclose(fp);
	printf("</table>\n");
	printf("<form method=get action=\"%s\">\n", SCRIPT);
	printf("<input type=hidden name=hostname value=\"%s\">\n", hostname);
	printf("<input type=hidden name=username value=\"%s\">\n", username);
	printf("<input type=text size=65 name=filename value=\"\">\n");
	printf("<textarea cols=65 rows=10 name=scripttext wrap=\"physical\">\n");
	printf("</textarea><p>\n");
	printf("<input type=submit name=mode value=\"Add script\">\n");
	printf("</form>\n");
	html_footer();
}

static void html_delscript(void)
{
	char *hostname = require_query("hostname");
	char *filename = require_query("filename");
	char b[4096];

	snprintf(b, sizeof b, "%s delscript %s %s", USH, hostname, filename);
	system(b);
	/* now list the result */
	html_script();
}

static void html_addscript(void)
{
	FILE *fp;
	char *hostname = require_query("hostname");
	char *filename = require_query("filename");
	char *scripttext = require_query("scripttext");
	char b[4096];

	fp = fopen(TMPFILE, "w");
	if (!fp) {
		html_error(TMPFILE);
		return;
	}
	fprintf(fp, "%s", scripttext);
	fclose(fp);
	snprintf(b, sizeof b, "%s script %s %s %s",
		USH, hostname, TMPFILE, filename);
	system(b);
	/* now list the result */
	html_script();
}

static struct {
	char *name;
	void (*fn)(void);
} fns[] = {
	{"CRONTAB", html_list},
	{"ADD", html_addjob},
	{"DELETE", html_delone},
	{"UPDATE", html_update},
	{"DELETE ALL", html_delall},
	{"UPLOAD", html_upload},
	{"DOWNLOAD", html_download},
	{"SCRIPTS", html_script},
	{"DELSCRIPT", html_delscript},
	{"ADD SCRIPT", html_addscript},
	{NULL, NULL}};

int main(void)
{
	int i;
	char *mode;
	if (parsecgi() == -1) {
		html_error("parsecgi");
		return 0;
	}
	mode = get_query("mode");
	sdb_init();
	url = getenv("UCRON_URL");
	if (!url) url = UCRON_URL;

#ifdef ORACLE_HOME
	if (getenv("ORACLE_HOME") == NULL) {
		putenv("ORACLE_HOME=" ORACLE_HOME);
	}
#endif

	if (mode == NULL) {
		html_default();
		return 0;
	}

	upcase(mode);
	for (i = 0; fns[i].name; i++) {
		if (!strcmp(mode, fns[i].name)) break;
	}
	if (fns[i].name) {
		(*fns[i].fn)();
	} else {
		cgi_err = mode;
		html_error("bogus mode");
	}
	return 0;
}

